@@ -12,24 +12,27 @@ from typing import (
     List,
     Dict,
     Iterable,
+    TYPE_CHECKING,
 )
 
 from debputy.analysis.analysis_util import flatten_ppfs
 from debputy.analysis.debian_dir import resolve_debhelper_config_files
 from debputy.dh.dh_assistant import extract_dh_compat_level
-from debputy.linting.lint_util import LintState
+from debputy.linting.lint_util import (
+    LintState,
+    te_range_to_lsp,
+    te_position_to_lsp,
+)
 from debputy.lsp.apt_cache import PackageLookup
 from debputy.lsp.debputy_ls import DebputyLanguageServer
-from debputy.lsp.diagnostics import DiagnosticData
 from debputy.lsp.lsp_debian_control_reference_data import (
     DctrlKnownField,
-    BINARY_FIELDS,
-    SOURCE_FIELDS,
     DctrlFileMetadata,
     package_name_to_section,
     all_package_relationship_fields,
     extract_first_value_and_position,
     all_source_relationship_fields,
+    StanzaMetadata,
 )
 from debputy.lsp.lsp_features import (
     lint_diagnostics,
@@ -40,8 +43,9 @@ from debputy.lsp.lsp_features import (
     lsp_semantic_tokens_full,
     lsp_will_save_wait_until,
     lsp_format_document,
-    LanguageDispatch,
     lsp_text_doc_inlay_hints,
+    LanguageDispatchRule,
+    SecondaryLanguage,
 )
 from debputy.lsp.lsp_generic_deb822 import (
     deb822_completer,
@@ -58,25 +62,15 @@ from debputy.lsp.quickfixes import (
     propose_insert_text_on_line_after_diagnostic_quick_fix,
     propose_remove_range_quick_fix,
 )
-from debputy.lsp.spellchecking import default_spellchecker
-from debputy.lsp.text_util import (
-    normalize_dctrl_field_name,
-    LintCapablePositionCodec,
-    te_range_to_lsp,
-    te_position_to_lsp,
-)
 from debputy.lsp.vendoring._deb822_repro import (
     Deb822FileElement,
     Deb822ParagraphElement,
 )
 from debputy.lsp.vendoring._deb822_repro.parsing import (
     Deb822KeyValuePairElement,
-    LIST_SPACE_SEPARATED_INTERPRETATION,
 )
 from debputy.lsprotocol.types import (
-    DiagnosticSeverity,
     Range,
-    Diagnostic,
     Position,
     FoldingRange,
     FoldingRangeParams,
@@ -104,6 +98,11 @@ from debputy.packager_provided_files import (
 from debputy.plugin.api.impl import plugin_metadata_for_debputys_own_plugin
 from debputy.util import detect_possible_typo, PKGNAME_REGEX, _info
 
+if TYPE_CHECKING:
+    pass
+else:
+    pass
+
 try:
     from debputy.lsp.vendoring._deb822_repro.locatable import (
         Position as TEPosition,
@@ -116,13 +115,16 @@ except ImportError:
     pass
 
 
-_LANGUAGE_IDS = [
-    LanguageDispatch.from_language_id("debian/control"),
-    # emacs's name
-    LanguageDispatch.from_language_id("debian-control"),
-    # vim's name
-    LanguageDispatch.from_language_id("debcontrol"),
-]
+_DISPATCH_RULE = LanguageDispatchRule.new_rule(
+    "debian/control",
+    "debian/control",
+    [
+        # emacs's name
+        SecondaryLanguage("debian-control"),
+        # vim's name
+        SecondaryLanguage("debcontrol"),
+    ],
+)
 
 
 @dataclasses.dataclass(slots=True, frozen=True)
@@ -294,7 +296,7 @@ _SUBSTVARS_DOC = _substvars_metadata(
             Note: This substvar only appears in `debhelper-compat (= 14)`, or
             with use of `debputy` (at an integration level, where `debputy`
             runs `dpkg-shlibdeps`), or when passing relevant options to
-            `dpkg-shlibdeps`  (often via `dh_shlibdeps`) such as `-dPre-Depends`.
+            `dpkg-shlibdeps` (often via `dh_shlibdeps`) such as `-dPre-Depends`.
     """
         ),
     ),
@@ -303,10 +305,10 @@ _SUBSTVARS_DOC = _substvars_metadata(
 _DCTRL_FILE_METADATA = DctrlFileMetadata()
 
 
-lsp_standard_handler(_LANGUAGE_IDS, TEXT_DOCUMENT_CODE_ACTION)
+lsp_standard_handler(_DISPATCH_RULE, TEXT_DOCUMENT_CODE_ACTION)
 
 
-@lsp_hover(_LANGUAGE_IDS)
+@lsp_hover(_DISPATCH_RULE)
 def _debian_control_hover(
     ls: "DebputyLanguageServer",
     params: HoverParams,
@@ -432,6 +434,18 @@ def _render_package_lookup(
                 f"This build-dependency will activate the `dh` sequence called `{sequence}`."
             )
             segments.append("")
+
+        elif (
+            known_field.name == "Build-Depends"
+            and name.startswith("debputy-plugin-")
+            and len(name) > 15
+        ):
+            plugin_name = name[15:]
+            segments.append(
+                f"This build-dependency will activate the `debputy` plugin called `{plugin_name}`."
+            )
+            segments.append("")
+
         segments.extend(
             [
                 f"Synopsis: {provider.synopsis}",
@@ -488,6 +502,23 @@ def _render_package_by_name(
 
         This build-dependency will activate the `dh` sequence called `{sequence}`.
 
+        """
+            )
+            + _disclaimer(is_empty)
+        )
+    if (
+        known_field.name == "Build-Depends"
+        and name.startswith("debputy-plugin-")
+        and len(name) > 15
+    ):
+        plugin_name = name[15:]
+        return (
+            textwrap.dedent(
+                f"""\
+        # {name}
+
+        This build-dependency will activate the `debputy` plugin called `{plugin_name}`.
+
         """
             )
             + _disclaimer(is_empty)
@@ -641,7 +672,7 @@ def _custom_hover(
     return dispatch(ls, known_field, line, word_at_position)
 
 
-@lsp_completer(_LANGUAGE_IDS)
+@lsp_completer(_DISPATCH_RULE)
 def _debian_control_completions(
     ls: "DebputyLanguageServer",
     params: CompletionParams,
@@ -649,7 +680,7 @@ def _debian_control_completions(
     return deb822_completer(ls, params, _DCTRL_FILE_METADATA)
 
 
-@lsp_folding_ranges(_LANGUAGE_IDS)
+@lsp_folding_ranges(_DISPATCH_RULE)
 def _debian_control_folding_ranges(
     ls: "DebputyLanguageServer",
     params: FoldingRangeParams,
@@ -657,7 +688,7 @@ def _debian_control_folding_ranges(
     return deb822_folding_ranges(ls, params, _DCTRL_FILE_METADATA)
 
 
-@lsp_text_doc_inlay_hints(_LANGUAGE_IDS)
+@lsp_text_doc_inlay_hints(_DISPATCH_RULE)
 def _doc_inlay_hint(
     ls: "DebputyLanguageServer",
     params: InlayHintParams,
@@ -690,7 +721,10 @@ def _doc_inlay_hint(
         )
         stanza_def = _DCTRL_FILE_METADATA.classify_stanza(stanza, stanza_no)
         for known_field in stanza_def.stanza_fields.values():
-            if not known_field.inherits_from_source or known_field.name in stanza:
+            if (
+                not known_field.inheritable_from_other_stanza
+                or known_field.name in stanza
+            ):
                 continue
 
             inherited_value = source_stanza.get(known_field.name)
@@ -720,25 +754,20 @@ def _doc_inlay_hint(
     return inlay_hints
 
 
-def _paragraph_representation_field(
-    paragraph: Deb822ParagraphElement,
-) -> Deb822KeyValuePairElement:
-    return next(iter(paragraph.iter_parts_of_type(Deb822KeyValuePairElement)))
-
-
 def _source_package_checks(
     stanza: Deb822ParagraphElement,
     stanza_position: "TEPosition",
+    stanza_metadata: StanzaMetadata[DctrlKnownField],
     lint_state: LintState,
-    diagnostics: List[Diagnostic],
 ) -> None:
     vcs_fields = {}
+    source_fields = _DCTRL_FILE_METADATA["Source"].stanza_fields
     for kvpair in stanza.iter_parts_of_type(Deb822KeyValuePairElement):
-        name = normalize_dctrl_field_name(kvpair.field_name.lower())
+        name = stanza_metadata.normalize_field_name(kvpair.field_name.lower())
         if (
             not name.startswith("vcs-")
             or name == "vcs-browser"
-            or name not in SOURCE_FIELDS
+            or name not in source_fields
         ):
             continue
         vcs_fields[name] = kvpair
@@ -746,25 +775,16 @@ def _source_package_checks(
     if len(vcs_fields) < 2:
         return
     for kvpair in vcs_fields.values():
-        kvpair_range_server_units = te_range_to_lsp(
-            kvpair.range_in_parent().relative_to(stanza_position)
-        )
-        diagnostics.append(
-            Diagnostic(
-                lint_state.position_codec.range_to_client_units(
-                    lint_state.lines, kvpair_range_server_units
-                ),
-                f'Multiple Version Control fields defined ("{kvpair.field_name}")',
-                severity=DiagnosticSeverity.Warning,
-                source="debputy",
-                data=DiagnosticData(
-                    quickfixes=[
-                        propose_remove_range_quick_fix(
-                            proposed_title=f'Remove "{kvpair.field_name}"'
-                        )
-                    ]
-                ),
-            )
+        lint_state.emit_diagnostic(
+            kvpair.range_in_parent().relative_to(stanza_position),
+            f'Multiple Version Control fields defined ("{kvpair.field_name}")',
+            "warning",
+            "Policy 5.6.26",
+            quickfixes=[
+                propose_remove_range_quick_fix(
+                    proposed_title=f'Remove "{kvpair.field_name}"'
+                )
+            ],
         )
 
 
@@ -772,21 +792,21 @@ def _binary_package_checks(
     stanza: Deb822ParagraphElement,
     stanza_position: "TEPosition",
     source_stanza: Deb822ParagraphElement,
-    representation_field_range: Range,
+    representation_field_range: "TERange",
     lint_state: LintState,
-    diagnostics: List[Diagnostic],
 ) -> None:
     package_name = stanza.get("Package", "")
     source_section = source_stanza.get("Section")
     section_kvpair = stanza.get_kvpair_element(("Section", 0), use_get=True)
     section: Optional[str] = None
+    section_range: Optional["TERange"] = None
     if section_kvpair is not None:
         section, section_range = extract_first_value_and_position(
             section_kvpair,
             stanza_position,
-            lint_state,
         )
-    else:
+
+    if section_range is None:
         section_range = representation_field_range
     effective_section = section or source_section or "unknown"
     package_type = stanza.get("Package-Type", "")
@@ -800,22 +820,19 @@ def _binary_package_checks(
             package_type_kvpair = stanza.get_kvpair_element(
                 "Package-Type", use_get=True
             )
-            package_type_range = None
+            package_type_range: Optional["TERange"] = None
             if package_type_kvpair is not None:
                 _, package_type_range = extract_first_value_and_position(
                     package_type_kvpair,
                     stanza_position,
-                    lint_state,
                 )
             if package_type_range is None:
                 package_type_range = representation_field_range
-            diagnostics.append(
-                Diagnostic(
-                    package_type_range,
-                    'The Package-Type should be "udeb" given the package name',
-                    severity=DiagnosticSeverity.Warning,
-                    source="debputy",
-                )
+            lint_state.emit_diagnostic(
+                package_type_range,
+                'The Package-Type should be "udeb" given the package name',
+                "warning",
+                "debputy",
             )
         guessed_section = "debian-installer"
         section_diagnostic_rationale = " since it is an udeb"
@@ -834,278 +851,21 @@ def _binary_package_checks(
                 )
             ]
         assert section_range is not None  # mypy hint
-        diagnostics.append(
-            Diagnostic(
-                section_range,
-                f'The Section should be "{component_prefix}{guessed_section}"{section_diagnostic_rationale}',
-                severity=DiagnosticSeverity.Warning,
-                source="debputy",
-                data=DiagnosticData(quickfixes=quickfix_data),
-            )
+        lint_state.emit_diagnostic(
+            section_range,
+            f'The Section should be "{component_prefix}{guessed_section}"{section_diagnostic_rationale}',
+            "warning",
+            "debputy",
+            quickfixes=quickfix_data,
         )
 
 
-def _diagnostics_for_paragraph(
-    deb822_file: Deb822FileElement,
-    stanza: Deb822ParagraphElement,
-    stanza_position: "TEPosition",
-    source_stanza: Deb822ParagraphElement,
-    known_fields: Mapping[str, DctrlKnownField],
-    other_known_fields: Mapping[str, DctrlKnownField],
-    is_binary_paragraph: bool,
-    doc_reference: str,
-    lint_state: LintState,
-    diagnostics: List[Diagnostic],
-) -> None:
-    representation_field = _paragraph_representation_field(stanza)
-    representation_field_range = representation_field.range_in_parent().relative_to(
-        stanza_position
-    )
-    representation_field_range = lint_state.position_codec.range_to_client_units(
-        lint_state.lines,
-        te_range_to_lsp(representation_field_range),
-    )
-    for known_field in known_fields.values():
-        if known_field.name in stanza:
-            continue
-
-        diagnostics.extend(
-            known_field.field_omitted_diagnostics(
-                deb822_file,
-                representation_field_range,
-                stanza,
-                stanza_position,
-                source_stanza,
-                lint_state,
-            )
-        )
-
-    if is_binary_paragraph:
-        _binary_package_checks(
-            stanza,
-            stanza_position,
-            source_stanza,
-            representation_field_range,
-            lint_state,
-            diagnostics,
-        )
-    else:
-        _source_package_checks(
-            stanza,
-            stanza_position,
-            lint_state,
-            diagnostics,
-        )
-
-    seen_fields: Dict[str, Tuple[str, str, Range, List[Range]]] = {}
-
-    for kvpair in stanza.iter_parts_of_type(Deb822KeyValuePairElement):
-        field_name_token = kvpair.field_token
-        field_name = field_name_token.text
-        field_name_lc = field_name.lower()
-        normalized_field_name_lc = normalize_dctrl_field_name(field_name_lc)
-        known_field = known_fields.get(normalized_field_name_lc)
-        field_value = stanza[field_name]
-        kvpair_range_te = kvpair.range_in_parent().relative_to(stanza_position)
-        field_range_te = kvpair.field_token.range_in_parent().relative_to(
-            kvpair_range_te.start_pos
-        )
-        field_position_te = field_range_te.start_pos
-        field_range_server_units = te_range_to_lsp(field_range_te)
-        field_range = lint_state.position_codec.range_to_client_units(
-            lint_state.lines,
-            field_range_server_units,
-        )
-        field_name_typo_detected = False
-        existing_field_range = seen_fields.get(normalized_field_name_lc)
-        if existing_field_range is not None:
-            existing_field_range[3].append(field_range)
-        else:
-            normalized_field_name = normalize_dctrl_field_name(field_name)
-            seen_fields[field_name_lc] = (
-                field_name,
-                normalized_field_name,
-                field_range,
-                [],
-            )
-
-        if known_field is None:
-            candidates = detect_possible_typo(normalized_field_name_lc, known_fields)
-            if candidates:
-                known_field = known_fields[candidates[0]]
-                token_range_server_units = te_range_to_lsp(
-                    TERange.from_position_and_size(
-                        field_position_te, kvpair.field_token.size()
-                    )
-                )
-                field_range = lint_state.position_codec.range_to_client_units(
-                    lint_state.lines,
-                    token_range_server_units,
-                )
-                field_name_typo_detected = True
-                diagnostics.append(
-                    Diagnostic(
-                        field_range,
-                        f'The "{field_name}" looks like a typo of "{known_field.name}".',
-                        severity=DiagnosticSeverity.Warning,
-                        source="debputy",
-                        data=DiagnosticData(
-                            quickfixes=[
-                                propose_correct_text_quick_fix(known_fields[m].name)
-                                for m in candidates
-                            ]
-                        ),
-                    )
-                )
-        if known_field is None:
-            known_else_where = other_known_fields.get(normalized_field_name_lc)
-            if known_else_where is not None:
-                intended_usage = "Source" if is_binary_paragraph else "Package"
-                diagnostics.append(
-                    Diagnostic(
-                        field_range,
-                        f'The {field_name} is defined for use in the "{intended_usage}" stanza.'
-                        f" Please move it to the right place or remove it",
-                        severity=DiagnosticSeverity.Error,
-                        source="debputy",
-                    )
-                )
-            continue
-
-        if field_value.strip() == "":
-            diagnostics.append(
-                Diagnostic(
-                    field_range,
-                    f"The {field_name} has no value. Either provide a value or remove it.",
-                    severity=DiagnosticSeverity.Error,
-                    source="debputy",
-                )
-            )
-            continue
-        diagnostics.extend(
-            known_field.field_diagnostics(
-                deb822_file,
-                kvpair,
-                stanza,
-                stanza_position,
-                kvpair_range_te,
-                lint_state,
-                field_name_typo_reported=field_name_typo_detected,
-            )
-        )
-        if known_field.spellcheck_value:
-            words = kvpair.interpret_as(LIST_SPACE_SEPARATED_INTERPRETATION)
-            spell_checker = default_spellchecker()
-            value_position = kvpair.value_element.position_in_parent().relative_to(
-                field_position_te
-            )
-            for word_ref in words.iter_value_references():
-                token = word_ref.value
-                for word, pos, endpos in spell_checker.iter_words(token):
-                    corrections = spell_checker.provide_corrections_for(word)
-                    if not corrections:
-                        continue
-                    word_loc = word_ref.locatable
-                    word_pos_te = word_loc.position_in_parent().relative_to(
-                        value_position
-                    )
-                    if pos:
-                        word_pos_te = TEPosition(0, pos).relative_to(word_pos_te)
-                    word_range_te = TERange(
-                        START_POSITION,
-                        TEPosition(0, endpos - pos),
-                    )
-                    word_range_server_units = te_range_to_lsp(
-                        TERange.from_position_and_size(word_pos_te, word_range_te)
-                    )
-                    word_range = lint_state.position_codec.range_to_client_units(
-                        lint_state.lines,
-                        word_range_server_units,
-                    )
-                    diagnostics.append(
-                        Diagnostic(
-                            word_range,
-                            f'Spelling "{word}"',
-                            severity=DiagnosticSeverity.Hint,
-                            source="debputy",
-                            data=DiagnosticData(
-                                lint_severity="spelling",
-                                quickfixes=[
-                                    propose_correct_text_quick_fix(c)
-                                    for c in corrections
-                                ],
-                            ),
-                        )
-                    )
-        source_value = source_stanza.get(field_name)
-        if known_field.warn_if_default and field_value == known_field.default_value:
-            diagnostics.append(
-                Diagnostic(
-                    field_range,
-                    f"The {field_name} is redundant as it is set to the default value and the field should only be"
-                    " used in exceptional cases.",
-                    severity=DiagnosticSeverity.Warning,
-                    source="debputy",
-                )
-            )
-
-        if known_field.inherits_from_source and field_value == source_value:
-            if range_compatible_with_remove_line_fix(field_range):
-                fix_data = propose_remove_line_quick_fix()
-            else:
-                fix_data = None
-            diagnostics.append(
-                Diagnostic(
-                    field_range,
-                    f"The field {field_name} duplicates the value from the Source stanza.",
-                    severity=DiagnosticSeverity.Information,
-                    source="debputy",
-                    data=DiagnosticData(quickfixes=fix_data),
-                )
-            )
-    for (
-        field_name,
-        normalized_field_name,
-        field_range,
-        duplicates,
-    ) in seen_fields.values():
-        if not duplicates:
-            continue
-        related_information = [
-            DiagnosticRelatedInformation(
-                location=Location(doc_reference, field_range),
-                message=f"First definition of {field_name}",
-            )
-        ]
-        related_information.extend(
-            DiagnosticRelatedInformation(
-                location=Location(doc_reference, r),
-                message=f"Duplicate of {field_name}",
-            )
-            for r in duplicates
-        )
-        for dup_range in duplicates:
-            diagnostics.append(
-                Diagnostic(
-                    dup_range,
-                    f"The {normalized_field_name} field name was used multiple times in this stanza."
-                    f" Please ensure the field is only used once per stanza. Note that {normalized_field_name} and"
-                    f" X[BCS]-{normalized_field_name} are considered the same field.",
-                    severity=DiagnosticSeverity.Error,
-                    source="debputy",
-                    related_information=related_information,
-                )
-            )
-
-
 def _scan_for_syntax_errors_and_token_level_diagnostics(
     deb822_file: Deb822FileElement,
-    position_codec: LintCapablePositionCodec,
-    lines: List[str],
-    diagnostics: List[Diagnostic],
+    lint_state: LintState,
 ) -> int:
-    first_error = len(lines) + 1
-    spell_checker = default_spellchecker()
+    first_error = len(lint_state.lines) + 1
+    spell_checker = lint_state.spellchecker()
     for (
         token,
         start_line,
@@ -1115,117 +875,113 @@ def _scan_for_syntax_errors_and_token_level_diagnostics(
     ) in deb822_token_iter(deb822_file.iter_tokens()):
         if token.is_error:
             first_error = min(first_error, start_line)
-            start_pos = Position(
+            start_pos = TEPosition(
                 start_line,
                 start_offset,
             )
-            end_pos = Position(
+            end_pos = TEPosition(
                 end_line,
                 end_offset,
             )
-            token_range = position_codec.range_to_client_units(
-                lines, Range(start_pos, end_pos)
-            )
-            diagnostics.append(
-                Diagnostic(
-                    token_range,
-                    "Syntax error",
-                    severity=DiagnosticSeverity.Error,
-                    source="debputy (python-debian parser)",
-                )
+            token_range = TERange.between(start_pos, end_pos)
+            lint_state.emit_diagnostic(
+                token_range,
+                "Syntax error",
+                "error",
+                "debputy",
             )
         elif token.is_comment:
             for word, col_pos, end_col_pos in spell_checker.iter_words(token.text):
                 corrections = spell_checker.provide_corrections_for(word)
                 if not corrections:
                     continue
-                start_pos = Position(
+                start_pos = TEPosition(
                     start_line,
                     col_pos,
                 )
-                end_pos = Position(
+                end_pos = TEPosition(
                     start_line,
                     end_col_pos,
                 )
-                word_range = position_codec.range_to_client_units(
-                    lines, Range(start_pos, end_pos)
-                )
-                diagnostics.append(
-                    Diagnostic(
-                        word_range,
-                        f'Spelling "{word}"',
-                        severity=DiagnosticSeverity.Hint,
-                        source="debputy",
-                        data=DiagnosticData(
-                            lint_severity="spelling",
-                            quickfixes=[
-                                propose_correct_text_quick_fix(c) for c in corrections
-                            ],
-                        ),
-                    )
+                word_range = TERange.between(start_pos, end_pos)
+                lint_state.emit_diagnostic(
+                    word_range,
+                    f'Spelling "{word}"',
+                    "spelling",
+                    "debputy",
+                    quickfixes=[propose_correct_text_quick_fix(c) for c in corrections],
+                    enable_non_interactive_auto_fix=False,
                 )
     return first_error
 
 
-@lint_diagnostics(_LANGUAGE_IDS)
-def _lint_debian_control(
-    lint_state: LintState,
-) -> Optional[List[Diagnostic]]:
-    lines = lint_state.lines
-    position_codec = lint_state.position_codec
+@lint_diagnostics(_DISPATCH_RULE)
+def _lint_debian_control(lint_state: LintState) -> None:
     doc_reference = lint_state.doc_uri
-    diagnostics = []
     deb822_file = lint_state.parsed_deb822_file_content
 
     first_error = _scan_for_syntax_errors_and_token_level_diagnostics(
         deb822_file,
-        position_codec,
-        lines,
-        diagnostics,
+        lint_state,
     )
 
-    paragraphs = list(deb822_file)
-    source_paragraph = paragraphs[0] if paragraphs else None
+    stanzas = list(deb822_file)
+    source_stanza = stanzas[0] if stanzas else None
     binary_stanzas_w_pos = []
 
-    for paragraph_no, paragraph in enumerate(paragraphs, start=1):
-        paragraph_pos = paragraph.position_in_file()
-        if paragraph_pos.line_position >= first_error:
+    source_stanza_metadata, binary_stanza_metadata = _DCTRL_FILE_METADATA.stanza_types()
+
+    for stanza_no, stanza in enumerate(stanzas, start=1):
+        stanza_position = stanza.position_in_file()
+        if stanza_position.line_position >= first_error:
             break
-        is_binary_paragraph = paragraph_no != 1
-        if is_binary_paragraph:
-            known_fields = BINARY_FIELDS
-            other_known_fields = SOURCE_FIELDS
-            binary_stanzas_w_pos.append((paragraph, paragraph_pos))
+        is_binary_stanza = stanza_no != 1
+        if is_binary_stanza:
+            stanza_metadata = binary_stanza_metadata
+            other_stanza_metadata = source_stanza_metadata
+            other_stanza_name = "Source"
+            binary_stanzas_w_pos.append((stanza, stanza_position))
+            _, representation_field_range = stanza_metadata.stanza_representation(
+                stanza, stanza_position
+            )
+            _binary_package_checks(
+                stanza,
+                stanza_position,
+                source_stanza,
+                representation_field_range,
+                lint_state,
+            )
         else:
-            known_fields = SOURCE_FIELDS
-            other_known_fields = BINARY_FIELDS
-        _diagnostics_for_paragraph(
+            stanza_metadata = source_stanza_metadata
+            other_stanza_metadata = binary_stanza_metadata
+            other_stanza_name = "Binary"
+            _source_package_checks(
+                stanza,
+                stanza_position,
+                stanza_metadata,
+                lint_state,
+            )
+
+        stanza_metadata.stanza_diagnostics(
             deb822_file,
-            paragraph,
-            paragraph_pos,
-            source_paragraph,
-            known_fields,
-            other_known_fields,
-            is_binary_paragraph,
+            stanza,
+            stanza_position,
             doc_reference,
             lint_state,
-            diagnostics,
+            confusable_with_stanza_metadata=other_stanza_metadata,
+            confusable_with_stanza_name=other_stanza_name,
+            inherit_from_stanza=source_stanza if is_binary_stanza else None,
         )
 
     _detect_misspelled_packaging_files(
         lint_state,
         binary_stanzas_w_pos,
-        diagnostics,
     )
 
-    return diagnostics
-
 
 def _package_range_of_stanza(
-    lint_state: LintState,
     binary_stanzas: List[Tuple[Deb822ParagraphElement, TEPosition]],
-) -> Iterable[Tuple[str, Optional[str], Range]]:
+) -> Iterable[Tuple[str, Optional[str], "TERange"]]:
     for stanza, stanza_position in binary_stanzas:
         kvpair = stanza.get_kvpair_element(("Package", 0), use_get=True)
         if kvpair is None:
@@ -1233,10 +989,6 @@ def _package_range_of_stanza(
         representation_field_range = kvpair.range_in_parent().relative_to(
             stanza_position
         )
-        representation_field_range = lint_state.position_codec.range_to_client_units(
-            lint_state.lines,
-            te_range_to_lsp(representation_field_range),
-        )
         yield stanza["Package"], stanza.get("Architecture"), representation_field_range
 
 
@@ -1323,11 +1075,9 @@ def _packaging_files(
 def _detect_misspelled_packaging_files(
     lint_state: LintState,
     binary_stanzas_w_pos: List[Tuple[Deb822ParagraphElement, TEPosition]],
-    diagnostics: List[Diagnostic],
 ) -> None:
     stanza_ranges = {
-        p: (a, r)
-        for p, a, r in _package_range_of_stanza(lint_state, binary_stanzas_w_pos)
+        p: (a, r) for p, a, r in _package_range_of_stanza(binary_stanzas_w_pos)
     }
     for ppf in _packaging_files(lint_state):
         binary_package = ppf.package_name
@@ -1348,56 +1098,40 @@ def _detect_misspelled_packaging_files(
         if likely_typo_of is not None:
             # Handles arch_restriction == 'all' at the same time due to how
             # the `likely-typo-of` is created
-            diagnostics.append(
-                Diagnostic(
-                    diag_range,
-                    f'The file "{path}" is likely a typo of "{likely_typo_of}"',
-                    severity=DiagnosticSeverity.Warning,
-                    source="debputy",
-                    data=DiagnosticData(
-                        report_for_related_file=path,
-                    ),
-                )
+            lint_state.emit_diagnostic(
+                diag_range,
+                f'The file "{path}" is likely a typo of "{likely_typo_of}"',
+                "warning",
+                "debputy",
+                diagnostic_applies_to_another_file=path,
             )
             continue
         if declared_arch == "all" and arch_restriction is not None:
-            diagnostics.append(
-                Diagnostic(
-                    diag_range,
-                    f'The file "{path}" has an architecture restriction but is for an `arch:all` package, so'
-                    f" the restriction does not make sense.",
-                    severity=DiagnosticSeverity.Warning,
-                    source="debputy",
-                    data=DiagnosticData(
-                        report_for_related_file=path,
-                    ),
-                )
+            lint_state.emit_diagnostic(
+                diag_range,
+                f'The file "{path}" has an architecture restriction but is for an `arch:all` package, so'
+                f" the restriction does not make sense.",
+                "warning",
+                "debputy",
+                diagnostic_applies_to_another_file=path,
             )
         elif arch_restriction == "all":
-            diagnostics.append(
-                Diagnostic(
-                    diag_range,
-                    f'The file "{path}" has an architecture restriction of `all` rather than a real architecture',
-                    severity=DiagnosticSeverity.Warning,
-                    source="debputy",
-                    data=DiagnosticData(
-                        report_for_related_file=path,
-                    ),
-                )
+            lint_state.emit_diagnostic(
+                diag_range,
+                f'The file "{path}" has an architecture restriction of `all` rather than a real architecture',
+                "warning",
+                "debputy",
+                diagnostic_applies_to_another_file=path,
             )
 
         if not ppf.definition.has_active_command:
-            diagnostics.append(
-                Diagnostic(
-                    diag_range,
-                    f"The file {path} is related to a command that is not active in the dh sequence"
-                    " with the current addons",
-                    severity=DiagnosticSeverity.Warning,
-                    source="debputy",
-                    data=DiagnosticData(
-                        report_for_related_file=path,
-                    ),
-                )
+            lint_state.emit_diagnostic(
+                diag_range,
+                f"The file {path} is related to a command that is not active in the dh sequence"
+                " with the current addons",
+                "warning",
+                "debputy",
+                diagnostic_applies_to_another_file=path,
             )
             continue
 
@@ -1412,21 +1146,18 @@ def _detect_misspelled_packaging_files(
                 or_alt_name = f' (or maybe "debian/{binary_package}.{basename}")'
             else:
                 or_alt_name = ""
-            diagnostics.append(
-                Diagnostic(
-                    diag_range,
-                    f'Possible typo in "{path}". Consider renaming the file to "debian/{alt_name}"'
-                    f"{or_alt_name} if it is intended for {binary_package}",
-                    severity=DiagnosticSeverity.Warning,
-                    source="debputy",
-                    data=DiagnosticData(
-                        report_for_related_file=path,
-                    ),
-                )
+
+            lint_state.emit_diagnostic(
+                diag_range,
+                f'Possible typo in "{path}". Consider renaming the file to "debian/{alt_name}"'
+                f"{or_alt_name} if it is intended for {binary_package}",
+                "warning",
+                "debputy",
+                diagnostic_applies_to_another_file=path,
             )
 
 
-@lsp_will_save_wait_until(_LANGUAGE_IDS)
+@lsp_will_save_wait_until(_DISPATCH_RULE)
 def _debian_control_on_save_formatting(
     ls: "DebputyLanguageServer",
     params: WillSaveTextDocumentParams,
@@ -1442,17 +1173,17 @@ def _reformat_debian_control(
     return deb822_format_file(lint_state, _DCTRL_FILE_METADATA)
 
 
-@lsp_format_document(_LANGUAGE_IDS)
-def _debian_control_on_save_formatting(
+@lsp_format_document(_DISPATCH_RULE)
+def _debian_control_format_file(
     ls: "DebputyLanguageServer",
     params: DocumentFormattingParams,
 ) -> Optional[Sequence[TextEdit]]:
     doc = ls.workspace.get_text_document(params.text_document.uri)
     lint_state = ls.lint_state(doc)
-    return deb822_format_file(lint_state, _DCTRL_FILE_METADATA)
+    return _reformat_debian_control(lint_state)
 
 
-@lsp_semantic_tokens_full(_LANGUAGE_IDS)
+@lsp_semantic_tokens_full(_DISPATCH_RULE)
 def _debian_control_semantic_tokens_full(
     ls: "DebputyLanguageServer",
     request: SemanticTokensParams,
