@@ -45,6 +45,8 @@ except ImportError:
     def detect_possible_typo(
         provided_value: str,
         known_values: Iterable[str],
+        *,
+        max_edit_distance: int = 2,
     ) -> Sequence[str]:
         return ()
 
@@ -55,14 +57,16 @@ else:
     def detect_possible_typo(
         provided_value: str,
         known_values: Iterable[str],
+        *,
+        max_edit_distance: int = 2,
     ) -> Sequence[str]:
         k_len = len(provided_value)
         candidates = []
         for known_value in known_values:
-            if abs(k_len - len(known_value)) > 2:
+            if abs(k_len - len(known_value)) > max_edit_distance:
                 continue
             d = distance(provided_value, known_value)
-            if d > 2:
+            if d > max_edit_distance:
                 continue
             candidates.append(known_value)
         return candidates
@@ -526,10 +530,20 @@ def scratch_dir(*, create_if_not_exists: bool = True) -> str:
     if create_if_not_exists:
         ensure_dir(_SCRATCH_DIR)
         if is_debputy_dir:
-            Path("debian/.debputy/.gitignore").write_text("*\n")
+            generated_root_directory("debian/.debputy", internal_only=True)
     return _SCRATCH_DIR
 
 
+def generated_root_directory(path: str, *, internal_only: bool = False) -> None:
+    root_dir = Path(path)
+    (root_dir / ".gitignore").write_text("*\n")
+    # TODO: Should we add a "CACHEDIR.TAG" here? (Requires a relevant ignore rule
+    if internal_only:
+        (root_dir / "CACHEDIR.TAG").write_bytes(
+            b"Signature: 8a477f597d28d172789f06886806bc55"
+        )
+
+
 _RUNTIME_CONTAINER_DIR_KEY: Optional[str] = None
 
 
@@ -569,7 +583,7 @@ def generated_content_dir(
 
 
 PerlConfigVars = collections.namedtuple(
-    "PerlIncDir", ["vendorlib", "vendorarch", "cross_inc_dir", "ld", "path_sep"]
+    "PerlConfigVars", ["vendorlib", "vendorarch", "cross_inc_dir", "ld", "path_sep"]
 )
 PerlConfigData = collections.namedtuple("PerlConfigData", ["version", "debian_abi"])
 _PERL_MODULE_DIRS: Dict[str, PerlConfigVars] = {}
@@ -611,21 +625,22 @@ def resolve_perl_config(
 ) -> PerlConfigVars:
     global _PERL_MODULE_DIRS
     if dpkg_architecture_variables.is_cross_compiling:
-        arch = (
-            dctrl_bin.resolved_architecture
-            if dctrl_bin is not None
-            else dpkg_architecture_variables.current_host_arch
-        )
+        if dctrl_bin is not None:
+            arch = dctrl_bin.resolved_architecture
+            ma = dctrl_bin.deb_multiarch
+        else:
+            arch = dpkg_architecture_variables.current_host_arch
+            ma = dpkg_architecture_variables.current_host_multiarch
     else:
+        # We are de facto using the build-arch version of perl here; be explicit
         arch = "_build_arch_"
+        ma = dpkg_architecture_variables["DEB_BUILD_MULTIARCH"]
     config_vars = _PERL_MODULE_DIRS.get(arch)
     if config_vars is None:
         cmd = ["perl"]
         if dpkg_architecture_variables.is_cross_compiling:
             version = _perl_version()
-            cross_inc_dir = (
-                f"/usr/lib/{dctrl_bin.deb_multiarch}/perl/cross-config-{version}"
-            )
+            cross_inc_dir = f"/usr/lib/{ma}/perl/cross-config-{version}"
             # FIXME: This should not fallback to "build-arch" but on the other hand, we use the perl module dirs
             #  for every package at the moment. So mandating correct perl dirs implies mandating perl-xs-dev in
             #  cross builds... meh.
@@ -863,6 +878,7 @@ def change_log_level(
 ) -> None:
     if _DEFAULT_LOGGER is not None:
         _DEFAULT_LOGGER.setLevel(log_level)
+    logging.getLogger("").setLevel(log_level)
 
 
 def current_log_level() -> Optional[int]:
